/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.client;

import net.craftforge.essential.controller.constants.Charset;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;

/**
 * A client response is used to provide a normalized response output for a client.
 *
 * @author Christian Bick
 * @since 04.07.11
 */
public class ClientResponse {

    private int statusCode;

    /**
     * The HTTP response body
     */
    private InputStream responseBody;

    /**
     * Creates a client response using the given HTTP status code.
     * The response is assumed not to contain a HTTP response body.
     *
     * @param statusCode The HTTP status code
     */
    public ClientResponse(int statusCode) {
        this.statusCode = statusCode;
    }

    /**
     * Creates a client response using the given HTTP status code
     * and the specified response body.
     *
     * @param statusCode The HTTP status code
     * @param responseBody The HTTP response body
     */
    public ClientResponse(int statusCode, InputStream responseBody) {
        this(statusCode);
        this.responseBody = responseBody;
    }

    /**
     * Gets the HTTP status code.
     *
     * @return The HTTP status code
     */
    public int getStatusCode() {
        return statusCode;
    }

    /**
     * Gets the HTTP response body.
     *
     * @return The HTTP response body
     */
    public InputStream getResponseBodyStream() {
        return responseBody;
    }

    /**
     * Gets the HTTP response body as <code>String</code> interpreting the
     * input to be encoded in UTF-8.
     *
     * @return The response body as <code>String</code>
     */
    public String getResponseBodyAsString() {
        return getResponseBodyAsString(Charset.UTF8);
    }

    /**
     * Gets the HTTP response body as <code>String</code> interpreting the
     * input to be encoded in the given charset.
     *
     * @param charset The charset encoding
     * @return The response body as <code>String</code>
     */
    public String getResponseBodyAsString(String charset) {
        try {
            StringBuilder sb = new StringBuilder();
            BufferedReader br = new BufferedReader(new InputStreamReader(responseBody, charset));
            int nextChar;
            while ((nextChar = br.read()) != -1) {
                sb.append((char)nextChar);
            }
            String responseBodyString = sb.toString();
            responseBody = new ByteArrayInputStream(responseBodyString.getBytes(charset));
            return responseBodyString;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
}
