/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.client;

import net.craftforge.essential.controller.constants.Charset;
import net.craftforge.essential.controller.constants.HttpHeader;
import net.craftforge.essential.controller.constants.MediaType;

import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;

/**
 * An HTTP client performs client request via HTTP using an HTTP URL connection.
 *
 * @author Christian Bick
 * @since 06.07.11
 */
public class HttpClient implements Client {

    private String urlBasePart;

    /**
     * <p>Creates an HTTP client that uses the given URL base part for its
     * requests. The URL base part looks like follows.</p>
     *
     * <p><b>schema:</b> <code>protocol://host:port/context</code></p>
     *
     * <p><b>example:</b> <code>http://api.craftforge.net:8080/core</code></p>
     *
     * <p>
     *     The URL base part is prefixed to every URL info part of a request. Together
     *     with an optional URL query part the complete request URL is assembled from
     *     these three parts.
     * </p>
     * @param urlBasePart The URL base part
     */
    public HttpClient(String urlBasePart) {
        this.urlBasePart = urlBasePart;
    }

    /**
     * Performs an HTTP request from the data provided by the given client
     * request and retrieves an HTTP response that as provided as client
     * response.
     *
     * @param request The client request
     * @return The client response.
     */
    public ClientResponse performRequest(ClientRequest request) {
        HttpURLConnection connection = getHttpUrlConnection(request);
        // Establish a connection to the resource and stream the request body
        InputStream requestBody = request.getRequestBody();
        try {
            connection.connect();
            if (requestBody != null) {
                int nextByte = requestBody.read();
                while (nextByte != -1) {
                    connection.getOutputStream().write(nextByte);
                    nextByte = requestBody.read();
                }
            }
            return getResponse(connection);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Gets an HTTP URL connection from the data provided by the client request.
     *
     * @param request The client request
     * @return The HTTP URL connection
     */
    private HttpURLConnection getHttpUrlConnection(ClientRequest request) {
        HttpURLConnection connection;
        try {
            // Build url query part using the request parameters
            String urlQueryPart = "";
            for (String paramName : request.getParameters().keySet()) {
                String[] paramValues = request.getParameters().get(paramName);
                for (String paramValue : paramValues) {
                    if (urlQueryPart.isEmpty()) {
                        urlQueryPart += "?";
                    } else {
                        urlQueryPart += "&";
                    }
                    urlQueryPart += URLEncoder.encode(paramName, Charset.UTF8);
                    urlQueryPart += "=";
                    urlQueryPart += URLEncoder.encode(paramValue, Charset.UTF8);
                }
            }
            // Build URL using concatenating base part, info part and query part
            URL url = new URL(urlBasePart + "/" + request.getUrlInfoPart() + urlQueryPart);

            // Get the HTTP URL connection using the URL
            connection = (HttpURLConnection)url.openConnection();

            // Set the HTTP method (e.g. POST)
            connection.setRequestMethod(request.getHttpMethod());
            connection.setRequestProperty(HttpHeader.ACCEPT, MediaType.XML);
            connection.setRequestProperty(HttpHeader.ACCEPT_CHARSET, Charset.UTF8);

            // Set the header values
            for (String headerName : request.getHeaders().keySet()) {
                String headerValue = request.getHeaders().get(headerName);
                connection.addRequestProperty(headerName, headerValue);
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        return connection;
    }

    /**
     * Gets a client request from the data provided by the HTTP URL connection.
     *
     * @param connection The HTTP URL connection
     * @return The client response
     * @throws IOException if the input stream of the HTTP URL connection cannot be retrieved
     */
    private ClientResponse getResponse(HttpURLConnection connection) throws IOException {
        return new ClientResponse(connection.getResponseCode(), connection.getInputStream());
    }

}
