/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller;

import java.util.HashMap;
import java.util.Map;

/**
 * A controller configuration.
 *
 * @author Christian Bick
 * @since 03.03.2011
 */
public class Configuration {

    private String defaultConsumer = "net.craftforge.essential.supply.jaxb.JaxbConsumer";
    private String defaultProducer = "net.craftforge.essential.supply.jaxb.JaxbProducer";
    private String phaseFlow = "net.craftforge.essential.controller.phases.flow.StandardPhaseFlow";
    private String authenticationProvider = null;
    private String loggingProvider = null;
    private Map<String, String> xmlDeserializationProperties = new HashMap<String, String>();
    private Map<String, String> jsonDeserializationProperties = new HashMap<String, String>();
    private Map<String, String> xmlSerializationProperties = new HashMap<String, String>();
    private Map<String, String> jsonSerializationProperties = new HashMap<String, String>();

    /**
     * Gets the class path of the default consumer.
     *
     * @return The class path of the default consumer
     */
    public String getDefaultConsumer() {
        return defaultConsumer;
    }

    /**
     * <p>Sets the class path of the default consumer. The default
     * consumer is used if a service or its service class is not
     * annotated with the @Consumer annotation.</p>
     * <p>
     *     <b>Default:</b> <code>net.craftforge.essential.supply.jaxb.JaxbConsumer</code>
     * </p>
     *
     * @param defaultConsumer The class path of the default consumer
     */
    public void setDefaultConsumer(String defaultConsumer) {
        this.defaultConsumer = defaultConsumer;
    }

    /**
     * Gets the class path of the default producer.
     *
     * @return The class path of the default producer.
     */
    public String getDefaultProducer() {
        return defaultProducer;
    }

    /**
     * Sets the class path of the default producer. The default
     * producer is used if a service or its service class is not
     * annotated with the @Producer annotation.
     * <p>
     *     <b>Default:</b> <code>net.craftforge.essential.supply.jaxb.JaxbProducer</code>
     * </p>
     *
     * @param defaultProducer The class path of the default producer.
     */
    public void setDefaultProducer(String defaultProducer) {
        this.defaultProducer = defaultProducer;
    }

    /**
     * Gets the class path of the phase flow defining the controller's phase execution logic.
     *
     * @return The class path of the phase flow
     */
    public String getPhaseFlow() {
        return phaseFlow;
    }

    /**
     * Sets the class path of the phase flow defining the controller's phase execution logic.
     * <p>
     *     <b>Default:</b> <code>net.craftforge.essential.controller.phases.flow.StandardPhaseFlow</code>
     * </p>
     *
     * @param phaseFlow The class path of the phase flow
     */
    public void setPhaseFlow(String phaseFlow) {
        this.phaseFlow = phaseFlow;
    }

    /**
     * Gets the class path of the logging provider.
     *
     * @return The logging provider
     */
    public String getLoggingProvider() {
        return loggingProvider;
    }

    /**
     * Sets the class path of the logging provider. If no logging provider is
     * specified, the logging phase will be skipped.
     *
     * @param loggingProvider The logging provider
     */
    public void setLoggingProvider(String loggingProvider) {
        this.loggingProvider = loggingProvider;
    }

    /**
     * Gets the class path of the authentication provider.
     *
     * @return The authentication provider
     */
    public String getAuthenticationProvider() {
        return authenticationProvider;
    }

    /**
     * Sets the class path of the authentication provider. If no authentication provider is
     * specified, the authentication phase will be skipped.
     *
     * @param authenticationProvider The authentication provider
     */
    public void setAuthenticationProvider(String authenticationProvider) {
        this.authenticationProvider = authenticationProvider;
    }

    /**
     * Gets the XML deserialization properties. Adding properties according to
     * specifications can be used to fine-tune the XML deserialization process.
     *
     * @return The XML deserialization properties as Map
     */
    public Map<String, String> getXmlDeserializationProperties() {
        return xmlDeserializationProperties;
    }

    /**
     * Sets the XML deserialization properties. Adding properties according to
     * specifications can be used to fine-tune the XML deserialization process.
     *
     * @param xmlDeserializationProperties The XML deserialization properties as Map
     */
    public void setXmlDeserializationProperties(Map<String, String> xmlDeserializationProperties) {
        this.xmlDeserializationProperties = xmlDeserializationProperties;
    }

    /**
     * Gets the JSON deserialization properties. Adding properties according to
     * specifications can be used to fine-tune the JSON deserialization process.
     *
     * @return The JSON deserialization properties as Map
     */
    public Map<String, String> getJsonDeserializationProperties() {
        return jsonDeserializationProperties;
    }

    /**
     * Sets the JSON deserialization properties. Adding properties according to
     * specifications can be used to fine-tune the JSON deserialization process.
     *
     * @param jsonDeserializationProperties The JSON deserialization properties as Map
     */
    public void setJsonDeserializationProperties(Map<String, String> jsonDeserializationProperties) {
        this.jsonDeserializationProperties = jsonDeserializationProperties;
    }

    /**
     * Gets the XML serialization properties. Adding properties according to
     * specifications can be used to fine-tune the XML serialization process.
     *
     * @return The XML serialization properties as Map
     */
    public Map<String, String> getXmlSerializationProperties() {
        return xmlSerializationProperties;
    }

    /**
     * Sets the XML serialization properties. Adding properties according to
     * specifications can be used to fine-tune the XML serialization process.
     *
     * @param xmlSerializationProperties The XML serialization properties as Map
     */
    public void setXmlSerializationProperties(Map<String, String> xmlSerializationProperties) {
        this.xmlSerializationProperties = xmlSerializationProperties;
    }

    /**
     * Gets the JSON serialization properties. Adding properties according to
     * specifications can be used to fine-tune the JSON serialization process.
     *
     * @return The JSON serialization properties as Map
     */
    public Map<String, String> getJsonSerializationProperties() {
        return jsonSerializationProperties;
    }

    /**
     * Sets the JSON serialization properties. Adding properties according to
     * specifications can be used to fine-tune the JSON serialization process.
     *
     * @param jsonSerializationProperties The JSON serialization properties as Map
     */
    public void setJsonSerializationProperties(Map<String, String> jsonSerializationProperties) {
        this.jsonSerializationProperties = jsonSerializationProperties;
    }
}
