/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller;

import net.craftforge.essential.controller.constants.Charset;
import net.craftforge.essential.controller.constants.HttpHeader;
import net.craftforge.essential.controller.constants.HttpMethod;
import net.craftforge.essential.controller.constants.MediaType;
import net.craftforge.essential.controller.managers.ServiceManager;
import net.craftforge.essential.controller.utils.ControllerUtils;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.Map;

/**
 * Data structure containing all external (request and configuration)
 * information for the phases.
 *
 * @author Christian Bick
 * @since 06.02.2011
 */
public class Setup {

    /**
     * The service manager
     */
    private ServiceManager serviceManager;

    /**
     * The controller configuration
     */
    private Configuration configuration;

    /**
     * The normalized HTTP request
     */
    private Request request;

    /**
     * The normalized HTTP response
     */
    private Response response;

    /**
     * The username from the HTTP authorization header
     */
    private String httpUsername;

    /**
     * The password from the HTTP authorization header
     */
    private String httpPassword;

    /**
     * Constructs the setup.
     *
     * @param serviceManager The service manager
     * @param configuration The controller configuration
     * @param request The normalized HTTP request
     * @param response The normalized HTTP response
     */
    public Setup(ServiceManager serviceManager,
                   Configuration configuration,
                   Request request,
                   Response response) {
        
        this.serviceManager = serviceManager;
        this.configuration = configuration;
        this.request = request;
        this.response = response;

        // Initialize HTTP authorization username and password
        String authorization = request.getHeaders().get(HttpHeader.AUTHORIZATION);
        String[] userPass = ControllerUtils.getAuthPair(authorization);
        httpUsername = userPass[0];
        httpPassword = userPass[1];
    }

    /**
     * Gets the mime type for result serialization. Will use the accepted mime type of the request if
     * not overridden in the URL meta part of the URL info part. (e.g. my/path.json)
     *
     * @return The mime type
     */
    public String getAcceptedMediaType() {
        String metaDataPart = getRequestUrlMetaPart();
        String acceptedMediaType = request.getHeaders().get(HttpHeader.ACCEPT);
        if (metaDataPart.contains(".xml")) {
            return MediaType.XML;
        } else if (metaDataPart.contains(".json")) {
            return MediaType.JSON;
        } else if (acceptedMediaType == null) {
                return MediaType.XML;
        } else {
            return acceptedMediaType;
        }
    }

    /**
     * Gets the encoding charset for result serialization. Will use the accepted charset of the request.
     *
     * @return The encoding charset
     */
    public String getAcceptedCharset() {
        String acceptedCharset = request.getHeaders().get(HttpHeader.ACCEPT_CHARSET);
        if (acceptedCharset != null && !acceptedCharset.isEmpty()) {
            String[] parts = acceptedCharset.split(",");
            for (String part : parts) {
                part = part.trim();
                if (!part.equals("*")) {
                    return part;
                }
            }
        }
        return Charset.UTF8;
    }

    /**
     * Gets the HTTP method for result serialization. Will use the HTTP method of the request if
     * not overridden in the URL meta part of the URL info part. (e.g. my/path.put)
     *
     * @return The HTTP method
     */
    public String getHttpMethod() {
        String metaDataPart = getRequestUrlMetaPart();
        if (metaDataPart.contains(".get")) {
            return HttpMethod.GET;
        } else if (metaDataPart.contains(".post")) {
            return HttpMethod.POST;
        } else if (metaDataPart.contains(".put")) {
            return HttpMethod.PUT;
        } else if (metaDataPart.contains(".delete")) {
            return HttpMethod.DELETE;
        } else if (metaDataPart.contains(".head")) {
            return HttpMethod.HEAD;
        } else if (metaDataPart.contains(".options")) {
            return HttpMethod.OPTIONS;
        } else if (metaDataPart.contains(".trace")) {
            return HttpMethod.TRACE;
        }
        return request.getHttpMethod().toUpperCase();
    }

    /**
     * Gets the content type of the request body.
     *
     * @return The content type of the request body
     */
    public String getContentType() {
        return request.getHeaders().get(HttpHeader.CONTENT_TYPE);
    }

    /**
     * Gets the service manager.
     *
     * @return The service manager
     */
    public ServiceManager getServiceManager() {
        return serviceManager;
    }

    /**
     * Gets the controller configuration.
     *
     * @return The controller configuration
     */
    public Configuration getConfiguration() {
        return configuration;
    }

    /**
     * Gets the request URL resource path by removing the
     * meta data part from the URL info part. This is everything in
     * the URL info part from the beginning until the first occurring dot
     * (excluded). (my/path.put.xml -> my/path)
     *
     * @return The request URL resource part
     */
    public String getRequestUrlResourcePart() {
        String pathInfo = getUri();
        if (pathInfo.contains(".")) {
            pathInfo = pathInfo.substring(0, pathInfo.indexOf("."));
        }
        return ControllerUtils.standardUri(pathInfo);
    }

    /**
     * Gets the request URL meta part. This is everything in
     * the URL info part from the first occurring dot
     * (included) till the end. (my/path.put.xml -> .put.xml)
     *
     * @return The request URL meta part
     */
    public String getRequestUrlMetaPart() {
        String pathInfo = getUri();
        if (! pathInfo.contains(".")) {
            return "";
        }
        return pathInfo.substring(pathInfo.indexOf(".")-1, pathInfo.length());
    }

    /**
     * Gets the decoded user from the authorization header.
     *
     * @return The decoded user
     */
    public String getHttpAuthUser() {
        return httpUsername;

    }

    /**
     * Gets the decoded password from the authorization header.
     * 
     * @return The decoded password
     */
    public String getHttpAuthPass() {
        return httpPassword;
    }

    /**
     * Gets the request URL info part. This is the part that
     * is used by the framework for all allocation, path
     * parameter and meta data related information gathering.
     *
     * @return The request URL info part
     */
    public String getUri() {
        return request.getUrlInfoPart();
    }

    /**
     * Gets the parameter map, containing the query and form
     * parameters of the request.
     *
     * @return The parameter map
     */
    public Map<String, String[]> getParameterMap() {
        return request.getParameters();
    }

    /**
     * Gets the body input stream. Will return null if HTTP method is not
     * PUT or POST and if the header 'Content-Type' is set to application/x-www-form-urlencoded.
     *
     * @return The input stream of the request body
     */
    public InputStream getRequestBodyInputStream() {
        return request.getBodyInputStream();
    }

    /**
     * Gets the output stream for streaming the serialized response body.
     *
     * @return The output stream
     */
    public OutputStream getResponseBodyOutputStream() {
        return response.getBodyOutputStream();
    }

    /**
     * Gets the normalized HTTP request.
     *
     * @return The HTTP request
     */
    public Request getRequest() {
        return request;
    }

    /**
     * Gets the normalized HTTP response.
     *
     * @return The HTTP response
     */
    public Response getResponse() {
        return response;
    }
}
