/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */


package net.craftforge.essential.controller.managers;

import net.craftforge.essential.controller.ControllerException;
import net.craftforge.essential.controller.annotations.DefaultValue;
import net.craftforge.essential.controller.annotations.Param;
import net.craftforge.essential.controller.constants.HttpStatusCode;
import net.craftforge.essential.controller.utils.AnnotationUtils;
import net.craftforge.reflection.managers.ClassManager;
import net.craftforge.reflection.utils.ReflUtils;

import java.io.*;
import java.lang.annotation.Annotation;
import java.lang.reflect.*;
import java.util.*;

/**
 * Handles parameter access.
 *
 * @author Christian Bick
 * @since 27.02.2011
 */
public class ParameterManager {

    private Object consumer;
    private Map<String, String[]> parameterMap;
    private InputStream inputStream;
    private String contentType;
    private String httpUser;
    private String httpPass;

    /**
     * Constructs a parameter manager from the given consumer and parameter map.
     *
     * @param consumer The consumer being used for deserializing parameter values
     * @param parameterMap The parameter map containing the parameters and their values
     * @param inputStream The input stream of the request body
     * @param contentType The content type of the request body
     * @param httpUser The HTTP basic user
     * @param httpPass the HTTP basic password
     */
    public ParameterManager(Object consumer, Map<String, String[]> parameterMap,
                            InputStream inputStream, String contentType, String httpUser, String httpPass) {
        this.consumer = consumer;
        this.parameterMap = parameterMap;
        this.inputStream = inputStream;
        this.contentType = contentType;
        this.httpUser = httpUser;
        this.httpPass = httpPass;
    }

    /**
     * Gets the input for a field annotated with @Param from the corresponding parameter.
     *
     * @param field The field to get the input for
     * @return The input for the field
     * @throws ControllerException Failed to estimate a field's input
     */
    public Object getInputForField(Field field) throws ControllerException {
        Object input = null;
        String paramName = AnnotationUtils.getParamFromProperty(field);

        if (paramName != null) {
            if (paramName.equals("[http-body]")) {
                input = getBodyInput(field.getType());
            } else if (paramName.equals("[http-user]")) {
                input = httpUser;
            } else if (paramName.equals("[http-pass]")) {
                input = httpPass;
            } else {
                paramName = paramName.isEmpty() ? field.getName() : paramName;
                Class<?> type = field.getType();
                Type genericType = field.getGenericType();

                String[] paramValues = parameterMap.get(paramName);
                if (paramValues == null) {
                    paramValues = AnnotationUtils.getDefaultValuesFromProperty(field);
                }
                if (paramValues != null && paramValues.length > 0) {
                    if (Collection.class.isAssignableFrom(type) && genericType instanceof ParameterizedType) {
                        Class<?> typeArgument = (Class<?>)((ParameterizedType)genericType).getActualTypeArguments()[0];
                        input = getCollectionInput(type, typeArgument, paramValues);
                    } else {
                        input = getSingleInput(type, paramValues[0]);
                    }
                }
            }
        }

        if (input == null) {
            throw new ControllerException("The field " + field.getName() + " in class "
                    + field.getDeclaringClass() + " is required to be initialized");
        }
        return input;
    }

    /**
     * Gets the input for a method's parameters annotated with @Param from the corresponding parameter.
     *
     * @param method The method to get the inputs for its parameters
     * @return The list of inputs (ordered by parameter order)
     * @throws ControllerException Failed to estimate a parameter's input
     */
    public List<Object> getInputForMethodParameters(Method method) throws ControllerException {
        int paramIterator = 0;
        Class<?>[] paramTypes = method.getParameterTypes();
        Type[] paramGenericTypes = method.getGenericParameterTypes();

        List<Object> parameterInputs = new ArrayList<Object>();

        for (Annotation[] annotations : ClassManager.getInstance(method).getMethodParameterAnnotations(method, Param.class)) {
            Class<?> type = paramTypes[paramIterator];
            Type genericType = paramGenericTypes[paramIterator];

            Object value = null;
            String[] defaultValue = null;
            for (Annotation ann : annotations) {
                if (ann.annotationType().equals(DefaultValue.class)) {
                    defaultValue = ((DefaultValue)ann).value();
                    break;
                }
            }

            for (Annotation ann : annotations) {
                if (ann.annotationType().equals(Param.class)) {
                    String paramName = ((Param)ann).value();
                    if (paramName.isEmpty()) {
                        throw new ControllerException("A parameter annotation (@Param) of method " + method.getName()
                                + " in class " + method.getDeclaringClass() + " misses the name specification");
                    }
                    if (paramName.equals("[http-body]")) {
                        value = getBodyInput(type);
                    } else if (paramName.equals("[http-user]")) {
                        value = httpUser;
                    } else if (paramName.equals("[http-pass]")) {
                        value = httpPass;
                    } else {
                        String[] paramValues = parameterMap.get(paramName);
                        if (paramValues == null && defaultValue != null) {
                            paramValues = defaultValue;
                        }
                        if (paramValues != null && paramValues.length > 0) {
                            if (Collection.class.isAssignableFrom(type) && genericType instanceof ParameterizedType) {
                                Class<?> typeArgument = (Class<?>)((ParameterizedType)genericType).getActualTypeArguments()[0];
                                value = getCollectionInput(type, typeArgument, paramValues);
                                break;
                            } else {
                                value = getSingleInput(type, paramValues[0]);
                                break;
                            }
                        }
                    }
                }
            }

            if (value == null) {
                throw new ControllerException("A parameter of method " + method.getName() + " in class "
                        + method.getDeclaringClass() + " is required to be initialized");
            }

            paramIterator++;
            parameterInputs.add(value);
        }
        return parameterInputs;
    }

    /**
     * Gets the input for a collection type.
     *
     * @param collectionType The collection type
     * @param genericType The collection's generic type parameter class
     * @param paramValues The request parameter values
     * @param <G> The generic type parameter
     * @return The collection initialized from the parameter values
     * @throws net.craftforge.essential.controller.ControllerException Failed to initialize the collection
     */
    @SuppressWarnings("unchecked")
    protected <G> Collection<G> getCollectionInput(Class<?> collectionType, Class<G> genericType, String[] paramValues) throws ControllerException {
        Collection<G> collection = ReflUtils.createCollection(collectionType, genericType);
        for (String paramValue : paramValues) {
            collection.add((G) getSingleInput(genericType, paramValue));
        }
        return collection;
    }

    /**
     * Gets the input for a single object type.
     *
     * @param type The desired type
     * @param paramValue The request parameter value
     * @return The object initialized from the parameter values
     * @throws ControllerException Failed to initialize the object
     */
    protected Object getSingleInput(Class<?> type, String paramValue) throws ControllerException {
        Method method;
        if (ReflUtils.isSimpleDataType(type)) {
            method = AnnotationUtils.getConsumerMethod(consumer, "text/plain");
        } else if (isXml(paramValue)) {
            method = AnnotationUtils.getConsumerMethod(consumer, "text/xml");
        } else if (isJson(paramValue)) {
            method = AnnotationUtils.getConsumerMethod(consumer, "text/json");
        } else {
            throw new ControllerException("Failed to consume a parameter value because no " +
                    "content-type was specified for not self-explaining content.", HttpStatusCode.BadRequest);
        }
        try {
            if (method.getParameterTypes()[1].equals(InputStream.class)) {
                return method.invoke(consumer, type, new ByteArrayInputStream(paramValue.getBytes()));
            } else {
                return method.invoke(consumer, type, paramValue);
            }
        } catch (InvocationTargetException e) {
            Throwable cause = e.getCause();
            if (cause instanceof ControllerException) {
                throw (ControllerException)cause;
            }
            throw new ControllerException("The consumer method failed during execution: " + method, cause);
        } catch (IllegalAccessException e) {
            throw new ControllerException("The consumer method could not be invoked: " +
                    "" + method, e);
        }
    }

    /**
     * Gets the input for a single object from the request body
     *
     * @param type The desired type
     * @return The object initialized from the request body
     * @throws ControllerException Failed to initialize the object
     */
    protected Object getBodyInput(Class<?> type) throws ControllerException {
        if (inputStream == null) {
            return null;
        }
        if (type.equals(InputStream.class)) {
            return inputStream;
        }
            Method method = null;
            if (contentType != null) {
                method = AnnotationUtils.getConsumerMethod(consumer, contentType);
            }

            if (method == null || method.getParameterTypes()[1].equals(String.class)) {
                BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(inputStream));
                StringBuilder stringBuilder = new StringBuilder();
                String line;
                try {
                    while ((line = bufferedReader.readLine()) != null) {
                        stringBuilder.append(line).append("\n");
                    }
                    bufferedReader.close();
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
                return getSingleInput(type, stringBuilder.toString());
            } else {
                try {
                    return method.invoke(consumer, type, inputStream);
                } catch (IllegalAccessException e) {
                    Throwable cause = e.getCause();
                    if (cause instanceof ControllerException) {
                        throw (ControllerException)cause;
                    }
                    throw new ControllerException("The consumer method failed during execution: " + method, cause);
                } catch (InvocationTargetException e) {
                    throw new ControllerException("The consumer method could not be invoked: " +
                        "" + method, e);
                }
            }
    }

    /**
     * Tests a parameter value to be in JSON format.
     *
     * @param paramValue The parameter value
     * @return True if the value is of JSON format, false otherwise
     */
    protected boolean isJson(String paramValue) {
        String text = paramValue.trim();
        return text.startsWith("{") && text.endsWith("}");
    }

    /**
     * Tests a parameter value to be in XML format.
     *
     * @param paramValue The parameter value
     * @return True if the value is of XML format, false otherwise
     */
    protected boolean isXml(String paramValue) {
        String text = paramValue.trim();
        return text.startsWith("<") && text.endsWith(">");
    }
}
