/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */


package net.craftforge.essential.controller.managers;

import net.craftforge.essential.controller.utils.AnnotationUtils;
import net.craftforge.essential.controller.utils.ControllerUtils;
import net.craftforge.essential.controller.utils.RegExUtils;
import net.craftforge.reflection.managers.ClassManager;
import net.craftforge.reflection.utils.ClassUtils;
import net.craftforge.reflection.utils.PackageUtils;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Handles matching of resource URLs against services.
 *
 * @author Christian Bick
 * @since 02.02.2011
 */
public class ServiceManager {

    /**
     * Map of service manager instances
     */
    private static ConcurrentMap<String, ServiceManager> instances = new ConcurrentHashMap<String, ServiceManager>();

    /**
     * Controls instantiation process to ensure that only one instance per package path is used.
     *
     * @param packagePath The package path
     * @return The service manager responsible for the given package path
     */
    public static ServiceManager getInstance(String packagePath) {
        instances.putIfAbsent(packagePath, new ServiceManager(packagePath));
        return instances.get(packagePath);
    }

    /**
     * Comparator for comparing patterns against each other.
     */
    private class PatternComparator implements Comparator<Pattern> {

        /**
         * Compares two patterns by the following criteria:<br>
         * <br>
         * First criteria: Amount of all literals outside of a group<br>
         * Second criteria: Amount of groups<br>
         * Third criteria: Amount of none-default groups<br>
         * <br>
         * A pattern is smaller then another one, if its amount specified by
         * criteria 1-3 is smaller the amount of the other one.
         *
         * @param first First pattern
         * @param second Second pattern
         * @return 1 if the first pattern is larger then the second one,
         * -1 if the first pattern is smaller then the other one and 0 
         * if they are equal
         */
        public int compare(Pattern first, Pattern second) {
            int result = compare(RegExUtils.countLiteralsWithoutGroups(first), RegExUtils.countLiteralsWithoutGroups(second));
            if (result != 0) {
                return result;
            }
            result = compare(RegExUtils.countAllGroups(first), RegExUtils.countAllGroups(second));
            if (result != 0) {
                return result;
            }
            return compare(RegExUtils.countAllGroupsWithoutDefault(first), RegExUtils.countAllGroupsWithoutDefault(second));
        }

        /**
         * Compares two integers by natural order.
         *
         * @param countFirst The first integer
         * @param countSecond The second integer
         * @return 1 if the first integer is larger then the second one,
         * -1 if the first integer is smaller then the other one and 0
         * if they are equal
         */
        private int compare(int countFirst, int countSecond) {
            if (countFirst > countSecond) {
                return 1;
            } else if (countFirst < countSecond) {
                return -1;
            } else {
                return 0;
            }
        }
    }

    private String packagePath;
    private Map<String, List<String>> pathParams;
    private Map<String, Class<?>> serviceClasses;
    private Map<String, Map<String, Method>> serviceMethods;

    /**
     * Initializes a service manager from a package path. Finds all service
     * classes in the package path, analyzes them and collects the service
     * methods and their request path parameters.
     *
     * @param packagePath The package path
     */
    private ServiceManager(String packagePath) {
        this.packagePath = packagePath;

        serviceClasses = new HashMap<String, Class<?>>();
        serviceMethods = new HashMap<String, Map<String, Method>>();
        pathParams = new HashMap<String, List<String>>();

        for (Class<?> clazz : PackageUtils.findClasses(packagePath)) {
            if (Modifier.isAbstract(clazz.getModifiers())) {
                continue;
            }
            String pathForClass = AnnotationUtils.getPathFromClass(clazz);
            if (pathForClass == null) {
                continue;
            }
            String classRegEx = RegExUtils.getRegExFromPath(pathForClass);
            serviceClasses.put(classRegEx, clazz);
            pathParams.put(classRegEx, RegExUtils.getParametersFromPath(pathForClass));

            for (Method method : ClassManager.getInstance(clazz).getCompleteClassHierarchyMethods()) {
                if (! Modifier.isPublic(method.getModifiers())) {
                    continue;
                }
                String httpMethod = AnnotationUtils.getHttpMethod(method);
                if (httpMethod == null) {
                    continue;
                }
                String completePath = pathForClass;
                String pathForMethod = AnnotationUtils.getPathFromMethod(method);
                if (pathForMethod != null) {
                    completePath += pathForMethod;
                }
                String methodRegEx = RegExUtils.getRegExFromPath(completePath);
                pathParams.put(methodRegEx, RegExUtils.getParametersFromPath(completePath));
                if (! serviceMethods.containsKey(methodRegEx)) {
                    serviceMethods.put(methodRegEx, new HashMap<String, Method>());
                }
                if (! serviceMethods.get(methodRegEx).containsKey(httpMethod)) {
                    serviceMethods.get(methodRegEx).put(httpMethod, method);
                }
            }
        }
    }

    /**
     * Gets the package path
     *
     * @return The package path
     */
    public String getPackagePath() {
        return packagePath;
    }

    /**
     * Gets the request path parameter map corresponding to the request path.
     *
     * @param path The request path
     * @return The request path parameter map
     */
    public Map<String, String[]> getPathParameterMap(String path) {
        path = ControllerUtils.standardUri(path);
        List<Pattern> matchedPatterns =  new ArrayList<Pattern>();
        for (String regEx : pathParams.keySet()) {
            Pattern pattern = Pattern.compile(regEx);
            Matcher matcher = pattern.matcher(path);
            if (! matcher.matches()) {
                continue;
            }
            if (matcher.groupCount() > 0) {
                String finalGroup = matcher.group(matcher.groupCount()-1);
                if (finalGroup.isEmpty() || finalGroup.equals("/")) {
                    continue;
                }
            }
            matchedPatterns.add(pattern);
        }
        if (matchedPatterns.isEmpty()) {
            return null;
        }
        Collections.sort(matchedPatterns, new PatternComparator());
        Collections.reverse(matchedPatterns);
        List<String> paramNames = pathParams.get(matchedPatterns.get(0).pattern());
        List<String[]> paramValues = RegExUtils.getValuesFromPath(path);
        if (paramNames.size() > paramValues.size()) {
            return null;
        }
        Map<String, String[]> pathParameterMap = new HashMap<String, String[]>();
        int count = 0;
        for (String param : paramNames) {
            if (! param.equals("{}")) {
                pathParameterMap.put(param, paramValues.get(count));
            }
            count++;
        }
        return pathParameterMap;
    }

    /**
     * Gets the service class corresponding to the request path.
     *
     * @param path The request path
     * @return The service class
     */
    public Class<?> getServiceClass(String path) {
        path = ControllerUtils.standardUri(path);
        List<Pattern> matchedPatterns =  new ArrayList<Pattern>();
        for (String regEx : serviceClasses.keySet()) {
            Pattern pattern = Pattern.compile(regEx);
            Matcher matcher = pattern.matcher(path);
            if (! matcher.matches()) {
                continue;
            }
            if (matcher.groupCount() > 0) {
                String finalGroup = matcher.group(matcher.groupCount()-1);
                if (finalGroup.isEmpty() || finalGroup.equals("/")) {
                    continue;
                }
            }
            matchedPatterns.add(pattern);
        }
        if (matchedPatterns.isEmpty()) {
            return null;
        }
        Collections.sort(matchedPatterns, new PatternComparator());
        Collections.reverse(matchedPatterns);
        return serviceClasses.get(matchedPatterns.get(0).pattern());
    }

    /**
     * Gets the service method corresponding to the HTTP method
     * and request path.
     *
     * @param httpMethod The HTTP method
     * @param path The request path
     * @param serviceClass The service class for this method
     * @return The service method
     */
    public Method getServiceMethod(String httpMethod, String path, Class<?> serviceClass) {
        path = ControllerUtils.standardUri(path);
        List<Pattern> matchedPatterns =  new ArrayList<Pattern>();
        for (String regEx : serviceMethods.keySet()) {
            Pattern pattern = Pattern.compile(regEx);
            Matcher matcher = pattern.matcher(path);
            if (! matcher.matches()) {
                continue;
            }
            if (matcher.groupCount() > 0) {
                String finalGroup = matcher.group(matcher.groupCount()-1);
                if (finalGroup.isEmpty() || finalGroup.equals("/")) {
                    continue;
                }
            }
            matchedPatterns.add(pattern);
        }
        if (matchedPatterns.isEmpty()) {
            return null;
        }
        Collections.sort(matchedPatterns, new PatternComparator());
        Collections.reverse(matchedPatterns);
        Method method = null;
        int index = 0;
        while (method == null && index < matchedPatterns.size()) {
            method = serviceMethods.get(matchedPatterns.get(index++).pattern()).get(httpMethod);     
        }
        return method == null ? null : ClassUtils.getFirstImplementation(serviceClass, method);
    }
    
}
