/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.phases;

import net.craftforge.essential.controller.ControllerException;
import net.craftforge.essential.controller.Phase;
import net.craftforge.essential.controller.Setup;
import net.craftforge.essential.controller.State;
import net.craftforge.essential.controller.constants.HttpStatusCode;
import net.craftforge.essential.controller.constants.HttpMethod;
import net.craftforge.essential.controller.managers.ServiceManager;

import java.lang.reflect.Method;

/**
 * Allocates and sets the service class and service method.
 *
 * @author Christian Bick
 * @since 17.01.2011
 */
public class AllocationPhase implements Phase {

    private State state;
    private Setup setup;

    /**
     * Creates an allocation phase from a controller state and setup.
     *
     * @param state The controller state
     * @param setup The controller setup
     */
    public AllocationPhase(State state, Setup setup) {
        this.state = state;
        this.setup = setup;
    }

    /**
     * {@inheritDoc}
     */
    public void run() throws ControllerException {
        try {
            if (state.getServiceMethod() == null) {
                // Obtain HTTP method, resource URL and service manager from setup
                String httpMethod = setup.getHttpMethod();
                String resourcePath = setup.getRequestUrlResourcePart();
                ServiceManager serviceManager = setup.getServiceManager();

                // Allocate service class
                Class<?> serviceClass = serviceManager.getServiceClass(resourcePath);

                // Service class must be allocated in any case
                if (serviceClass == null) {
                    throw new ControllerException("'" + httpMethod + " " + resourcePath + "' has no responsible service class in module '" + serviceManager.getPackagePath() + "'", HttpStatusCode.NotFound);
                }
                state.setServiceClass(serviceClass);

                // Allocate service method
                Method serviceMethod = serviceManager.getServiceMethod(httpMethod, resourcePath, serviceClass);

                // Service method must be allocated for all HTTP methods except 'OPTIONS'
                if (serviceMethod == null && ! httpMethod.equals(HttpMethod.OPTIONS)) {
                    throw new ControllerException("'" + httpMethod + " " + resourcePath + "' has no responsible service method in module '" + serviceManager.getPackagePath() + "'", HttpStatusCode.NotFound);
                }
                state.setServiceMethod(serviceMethod);
            }
        } catch (RuntimeException e) {
            throw new ControllerException("An unexpected exception occurred during service allocation", e);
        }
    }
}
