/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.phases;

import net.craftforge.essential.controller.ControllerException;
import net.craftforge.essential.controller.Phase;
import net.craftforge.essential.controller.Setup;
import net.craftforge.essential.controller.State;
import net.craftforge.essential.controller.constants.HttpStatusCode;
import net.craftforge.essential.controller.providers.AuthenticationProvider;

/**
 * Authenticates the requester via HTTP basic authorization.
 *
 * @author Christian Bick
 * @since 22.05.2011
 */
public class AuthenticationPhase implements Phase {

    private State state;
    private Setup setup;

    /**
     * Creates an authentication phase from a controller state and setup.
     *
     * @param state The controller state
     * @param setup The controller setup
     */
    public AuthenticationPhase(State state, Setup setup) {
        this.state = state;
        this.setup = setup;
    }

    /**
     * {@inheritDoc}
     */
    public void run() throws ControllerException {
        String classPath = null;
        try {
            classPath = setup.getConfiguration().getAuthenticationProvider();
            if (classPath == null || classPath.isEmpty()) {
                return;
            }

            AuthenticationProvider provider;
            try {
                provider = (AuthenticationProvider)Class.forName(classPath).newInstance();
            } catch (ClassNotFoundException e) {
                throw new ControllerException("The authentication provider class specified in the controller" +
                        " configuration does not exist: " + classPath, e);
            } catch (InstantiationException e) {
                throw new ControllerException("The authentication provider class specified in the controller" +
                        " configuration could not be instantiated: " + classPath, e);
            } catch (IllegalAccessException e) {
                throw new ControllerException("The authentication provider class specified in the controller" +
                        " configuration could not be instantiated: " + classPath, e);
            }

            if (! provider.authenticates(setup.getHttpAuthUser(), setup.getHttpAuthPass())) {
                throw new ControllerException("Authentication failed because the user does not exist or" +
                        " user and password do not match", HttpStatusCode.Unauthorized);
            }

        } catch (RuntimeException e) {
            throw new ControllerException("An unexpected exception occurred during authentication", e);
        }
    }
}
