/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.phases;

import net.craftforge.essential.controller.ControllerException;
import net.craftforge.essential.controller.Phase;
import net.craftforge.essential.controller.Setup;
import net.craftforge.essential.controller.State;
import net.craftforge.essential.controller.annotations.DefaultValue;
import net.craftforge.essential.controller.annotations.Param;
import net.craftforge.essential.controller.constants.HttpStatusCode;
import net.craftforge.essential.controller.documentation.ServiceClassDoc;
import net.craftforge.essential.controller.documentation.ServiceMethodDoc;
import net.craftforge.essential.controller.documentation.ServiceParamDoc;
import net.craftforge.essential.controller.utils.AnnotationUtils;
import net.craftforge.reflection.managers.ClassManager;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.*;

/**
 * Generates the documentation of the service class and sets it as the result.
 *
 * @author Christian Bick
 * @since 06.03.2011
 */
public class DocumentationPhase implements Phase {

    private State state;
    private Setup setup;

    /**
     * Creates a documentation phase from a controller state and setup.
     *
     * @param state The controller state
     * @param setup The controller setup
     */
    public DocumentationPhase(State state, Setup setup) {
        this.state = state;
        this.setup = setup;
    }

    /**
     * {@inheritDoc}
     */
    public void run() throws ControllerException {
        try {
            String providerPath = AnnotationUtils.getPathFromClass(state.getServiceClass());
            ServiceClassDoc serviceClassDoc = new ServiceClassDoc(providerPath, new Date());

            List<ServiceParamDoc> serviceProviderParamDocs = new ArrayList<ServiceParamDoc>();

            for (Field field : AnnotationUtils.getParamFieldsFromClass(state.getServiceClass())) {
                String value = AnnotationUtils.getParamFromProperty(field);
                serviceProviderParamDocs.add(
                        new ServiceParamDoc(
                                value != null && ! value.isEmpty() ? value : field.getName(),
                                field.getType().getSimpleName(),
                                defaultValuesAsString(AnnotationUtils.getDefaultValuesFromProperty(field))
                        )
                );
            }

            List<ServiceMethodDoc> serviceMethodDocs = new ArrayList<ServiceMethodDoc>();

            for (Method method : AnnotationUtils.getServiceMethodsFromClass(state.getServiceClass())) {
                String servicePath = AnnotationUtils.getPathFromMethod(method);
                servicePath = servicePath == null ? "" : servicePath;
                String serviceUrl = providerPath + servicePath;
                String simpleServiceUrl = serviceUrl
                        + "." + AnnotationUtils.getHttpMethod(method).toLowerCase()
                        + "." + setup.getAcceptedMediaType().replaceAll(".*/", "");

                ServiceMethodDoc serviceMethodDoc = new ServiceMethodDoc(
                        AnnotationUtils.getHttpMethod(method),
                        servicePath,
                        serviceUrl,
                        simpleServiceUrl,
                        method.getReturnType().getSimpleName()
                );

                List<ServiceParamDoc> serviceParamDocs = new ArrayList<ServiceParamDoc>();
                int paramIterator = 0;
                Class<?>[] paramTypes = method.getParameterTypes();

                serviceParamDocs.addAll(serviceProviderParamDocs);

                for (Annotation[] annotations : ClassManager.getInstance(method).getMethodParameterAnnotations(method, Param.class)) {
                    Class<?> paramType = paramTypes[paramIterator++];
                    String defaultValue = null;
                    for (Annotation ann : annotations) {
                        if (ann.annotationType().equals(DefaultValue.class)) {
                            defaultValue = defaultValuesAsString(((DefaultValue)ann).value());
                        }
                    }

                    for (Annotation ann : annotations) {
                        if (ann.annotationType().equals(Param.class)) {
                            serviceParamDocs.add(
                                    new ServiceParamDoc(
                                            ((Param)ann).value(),
                                            paramType.getSimpleName(),
                                            defaultValue
                                    )
                            );
                        }
                    }
                }

                if (! serviceParamDocs.isEmpty()) {
                    serviceMethodDoc.setParameters(serviceParamDocs);
                }
                serviceMethodDocs.add(serviceMethodDoc);
            }

            if (! serviceMethodDocs.isEmpty()) {
                serviceClassDoc.setServices(serviceMethodDocs);
            }
            state.setResult(serviceClassDoc);
            state.setStatus(HttpStatusCode.OK);
        } catch (RuntimeException e) {
            throw new ControllerException("An unexpected exception occurred during service documentation");
        }
    }
    
    protected String defaultValuesAsString(String[] defaultValues) {
        String defaultValue = "";
        if (defaultValues == null) {
            return null;
        }
        if (defaultValues.length == 1) {
            return defaultValues[0];
        }
        defaultValue += "{";
        for (String value : defaultValues) {
            defaultValue += value + ", ";
        }
        defaultValue = defaultValue.substring(0, defaultValue.length()-2);
        defaultValue += "}";
        return defaultValue;
    }
}
