/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.phases;

import net.craftforge.essential.controller.*;
import net.craftforge.essential.controller.constants.HttpStatusCode;
import net.craftforge.essential.controller.managers.ParameterManager;
import net.craftforge.essential.controller.utils.AnnotationUtils;
import net.craftforge.reflection.utils.PropertyUtils;
import net.craftforge.reflection.utils.ReflUtils;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Creates an instance of the service class, initializes annotated class properties
 * and invokes the service method. Sets the result of the service method.
 *
 * @author Christian Bick
 * @since 03.02.2011
 */
public class InvocationPhase implements Phase {

    private State state;
    private Setup setup;

    /**
     * Creates an invocation phase from a controller state and setup.
     *
     * @param state The controller state
     * @param setup The controller setup
     */
    public InvocationPhase(State state, Setup setup) {
        this.state = state;
        this.setup = setup;
    }

    /**
     * {@inheritDoc}
     */
    public void run() throws ControllerException {
        try {
            // Initialize consumer
            String consumerPath = AnnotationUtils.getConsumerFromMethodOrClass(state.getServiceMethod());
            if (consumerPath == null || consumerPath.isEmpty()) {
                consumerPath = setup.getConfiguration().getDefaultConsumer();   
            }

            Class<?> consumerClass;
            try {
                consumerClass = Class.forName(consumerPath);
            } catch (ClassNotFoundException e) {
                throw new ControllerException("The consumer class specified in the controller" +
                        " configuration or annotated at the service method does not exist: " + consumerPath, e);
            }

            Object consumer;
            try {
                try {
                    consumer = consumerClass.getConstructor(Configuration.class).newInstance(setup.getConfiguration());
                } catch (NoSuchMethodException e) {
                    consumer = consumerClass.newInstance();
                }
            } catch (InstantiationException e) {
                throw new ControllerException("The consumer class specified in the controller" +
                        " configuration or annotated at the service method could not be" +
                        " instantiated: " + consumerPath, e);
            } catch (IllegalAccessException e) {
                throw new ControllerException("The consumer class specified in the controller" +
                        " configuration or annotated at the service method could not be" +
                        " instantiated: " + consumerPath, e);
            } catch (InvocationTargetException e) {
                throw new ControllerException("The consumer class specified in the controller" +
                        " configuration or annotated at the service method could not be" +
                        " instantiated: " + consumerPath, e);
            }
            
            // Union common parameter map with path parameter map
            Map<String, String[]> parameterMap = new HashMap<String, String[]>(setup.getParameterMap());
            Map<String, String[]> pathParameterMap = setup.getServiceManager().getPathParameterMap(setup.getUri());
            if (pathParameterMap != null) {
                 parameterMap.putAll(pathParameterMap);
            }

            // Initialize ParameterManager
            ParameterManager paramManager = new ParameterManager(consumer, parameterMap,
                    setup.getRequestBodyInputStream(), setup.getContentType(),
                    setup.getHttpAuthUser(), setup.getHttpAuthPass());

            // Initialize service class
            Object service;
            try {
                service = state.getServiceClass().newInstance();
            } catch (InstantiationException e) {
                Throwable cause = e.getCause();
                if (cause instanceof ControllerException) {
                    throw (ControllerException)cause;
                }
                throw new ControllerException("The allocated service class could not be" +
                        " instantiated: " + state.getServiceClass(), cause);
            } catch (IllegalAccessException e) {
                throw new ControllerException("The allocated service class could not be" +
                        " instantiated: " + state.getServiceClass(), e);
            }

            for (Field field : AnnotationUtils.getParamFieldsFromClass(service.getClass())) {
                try {
                    PropertyUtils.setProperty(service, field, paramManager.getInputForField(field));
                } catch (NoSuchMethodException e) {
                    throw new ControllerException("A property of the service class could not be" +
                            " set via the setter method: " + ReflUtils.setterName(field.getName()) + "()" +
                            " in service class " + state.getServiceClass().getName(), e);
                } catch (InvocationTargetException e) {
                    Throwable cause = e.getCause();
                    if (cause instanceof ControllerException) {
                        throw (ControllerException)cause;
                    }
                    throw new ControllerException("A property of the service class could not be" +
                            " set via the setter method: " + ReflUtils.setterName(field.getName()) + "()" +
                            " in service class " + state.getServiceClass(), cause);
                } catch (IllegalAccessException e) {
                    throw new ControllerException("A property of the service class could not be" +
                            " set: " + field.getName() +
                            " in service class " + state.getServiceClass(), e);
                }
            }

            // Invoke service method
            List<Object> parameterValues = paramManager.getInputForMethodParameters(state.getServiceMethod());
            Object result = null;
            state.setInvocationStart(new Date());
            try {
                if (! state.getServiceMethod().getReturnType().equals(void.class)) {
                    if (parameterValues.size() == 0) {
                        result = state.getServiceMethod().invoke(service);
                    } else {
                        result = state.getServiceMethod().invoke(service, parameterValues.toArray());
                    }
                    state.setStatus(HttpStatusCode.OK);
                } else {
                    if (parameterValues.size() == 0) {
                        state.getServiceMethod().invoke(service);
                    } else {
                        state.getServiceMethod().invoke(service, parameterValues.toArray());
                    }
                    state.setStatus(HttpStatusCode.NoContent);
                }
            } catch (InvocationTargetException e) {
                Throwable cause = e.getCause();
                if (cause instanceof ControllerException) {
                    throw (ControllerException)cause;
                }
                throw new ControllerException("The service method failed during execution: " +
                        "" + state.getServiceMethod(), cause);
            } catch (IllegalAccessException e) {
                throw new ControllerException("The service method could not be invoked: " +
                        "" + state.getServiceMethod(), e);
            }
            state.setInvocationEnd(new Date());
            state.setResult(result);
        } catch (RuntimeException e) {
            throw new ControllerException("An unexpected exception occurred during service invocation", e);
        }
    }
}
