/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.phases;

import net.craftforge.essential.controller.*;
import net.craftforge.essential.controller.constants.HttpHeader;
import net.craftforge.essential.controller.constants.HttpMethod;
import net.craftforge.essential.controller.utils.AnnotationUtils;

import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

/**
 * Produces a serialized response from the result.
 *
 * @author Christian Bick
 * @since 17.01.2011
 */
public class ProductionPhase implements Phase {

    private State state;
    private Setup setup;

    /**
     * Creates a production phase from a controller state and setup.
     *
     * @param state The controller state
     * @param setup The controller setup
     */
    public ProductionPhase(State state, Setup setup) {
        this.state = state;
        this.setup = setup;
    }

    /**
     * {@inheritDoc}
     */
    public void run() throws ControllerException {
        try {

            // Initialize production parameters
            Object result = state.getResult();
            OutputStream outputStream = setup.getResponseBodyOutputStream();
            String encoding = setup.getAcceptedCharset();

            // Initialize producer
            String producerPath = setup.getHttpMethod().equals(HttpMethod.OPTIONS)
                    ? setup.getConfiguration().getDefaultProducer()
                    : AnnotationUtils.getProducerFromMethodOrClass(state.getServiceMethod());
            if (producerPath == null || producerPath.isEmpty()) {
                producerPath = setup.getConfiguration().getDefaultProducer();
            }

            Class<?> producerClass;
            try {
                producerClass = Class.forName(producerPath);
            } catch (ClassNotFoundException e) {
                throw new ControllerException("The producer class specified in the controller" +
                        " configuration or annotated at the service method could not be" +
                        " instantiated: " + producerPath, e);
            }

            Object producer;
            try {
                try {
                    producer = producerClass.getConstructor(Configuration.class).newInstance(setup.getConfiguration());
                } catch (NoSuchMethodException e) {
                    producer = producerClass.newInstance();
                }
            } catch (InstantiationException e) {
                throw new ControllerException("The producer class specified in the controller" +
                        " configuration or annotated at the service method could not be" +
                        " instantiated: " + producerPath, e);
            } catch (IllegalAccessException e) {
                throw new ControllerException("The producer class specified in the controller" +
                        " configuration or annotated at the service method could not be" +
                        " instantiated: " + producerPath, e);
            } catch (InvocationTargetException e) {
                throw new ControllerException("The producer class specified in the controller" +
                        " configuration or annotated at the service method could not be" +
                        " instantiated: " + producerPath, e);
            }

            // Send headers
            Map<String, String> headers = new HashMap<String, String>();
            headers.put(HttpHeader.CONTENT_TYPE, setup.getAcceptedMediaType() + "; charset=" + setup.getAcceptedCharset());
            setup.getResponse().sendHeaders(state.getStatus(), headers);

            // Invoke producer method
            Method producerMethod = AnnotationUtils.getProducerMethod(producer, setup.getAcceptedMediaType());
            Class<?>[] parameterTypes = producerMethod.getParameterTypes();
            try {
                if (Arrays.equals(parameterTypes, new Class<?>[]{Object.class, OutputStream.class, String.class})) {
                    producerMethod.invoke(producer, result, outputStream, encoding);
                } else {
                    throw new ControllerException("Producer method does not have a valid" +
                            " signature: " + producerMethod);
                }
            } catch (InvocationTargetException e) {
                Throwable cause = e.getCause();
                if (cause instanceof ControllerException) {
                    throw (ControllerException)cause;
                }
                throw new ControllerException("The producer method failed during execution: " + producerMethod, cause);
            } catch (IllegalAccessException e) {
                throw new ControllerException("The producer method could not be invoked: " +
                        "" + producerMethod, e);
            }
        } catch (RuntimeException e) {
            throw new ControllerException("An unexpected exception occurred during production", e);
        }
    }
}
