/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.phases.flow;

import net.craftforge.essential.controller.ControllerException;
import net.craftforge.essential.controller.Phase;
import net.craftforge.essential.controller.Setup;
import net.craftforge.essential.controller.State;
import net.craftforge.essential.controller.constants.HttpMethod;
import net.craftforge.essential.controller.phases.*;
import net.craftforge.essential.controller.utils.AnnotationUtils;

/**
 * The standard phase flow defines a default flow for phases.
 *
 * @author Christian Bick
 * @since 30.05.2011
 */
public class StandardPhaseFlow implements PhaseFlow {

    private State state;
    private Setup setup;

    /**
     * {@inheritDoc}
     */
    public void setState(State state) {
        this.state = state;
    }

    /**
     * {@inheritDoc}
     */
    public void setSetup(Setup setup) {
        this.setup = setup;
    }

    /**
     * {@inheritDoc}
     */
    public Phase getFirstPhase() {
        return new AllocationPhase(state, setup);
    }

    /**
     * {@inheritDoc}
     */
    public Phase getPhaseAfter(Phase phase) throws ControllerException {
        if (phase instanceof AllocationPhase) {
            return getPhaseAfterAllocation();
        } else if (phase instanceof AuthenticationPhase) {
            return getPhaseAfterAuthentication();
        } else if (phase instanceof InvocationPhase) {
            return getPhaseAfterInvocation();
        } else if (phase instanceof DocumentationPhase) {
            return getPhaseAfterDocumentation();
        } else if (phase instanceof ProductionPhase) {
            return getPhaseAfterProduction();
        } else if (phase instanceof LoggingPhase) {
            return getPhaseAfterLogging();
        } else {
            throw new ControllerException("Undefined flow for phase: " + phase.getClass());
        }
    }

    /**
     * Gets the phase following an allocation phase. If the HTTP method is OPTIONS
     * the documentation phase will follow. Otherwise, if an authentication provider
     * is available the next phase is the Authentication
     *
     * @return
     */
    protected Phase getPhaseAfterAllocation() {
        // Next phase is documentation phase for HTTP method 'OPTIONS'
        if (setup.getHttpMethod().equals(HttpMethod.OPTIONS)) {
            return new DocumentationPhase(state, setup);
        }
        // Next phase is authentication phase if an authentication provider is given in the config and the service is
        // not annotated to be public
        if (setup.getConfiguration().getAuthenticationProvider() != null
                && ! AnnotationUtils.isPublic(state.getServiceMethod())) {
            return new AuthenticationPhase(state, setup);
        }
        // Next phase is invocation phase for all other methods
        return new InvocationPhase(state, setup);
    }

    protected Phase getPhaseAfterAuthentication() {
        return new InvocationPhase(state, setup);
    }

    protected Phase getPhaseAfterInvocation() {
        if (state.getResult() == null) {
            return new LoggingPhase(state, setup);
        }
        return new ProductionPhase(state, setup);
    }

    protected Phase getPhaseAfterDocumentation() {
        return new ProductionPhase(state, setup);
    }

    protected Phase getPhaseAfterProduction() {
        return new LoggingPhase(state, setup);
    }

    protected Phase getPhaseAfterLogging() {
        return null;
    }
}
