/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.utils;

import net.craftforge.essential.controller.ControllerException;
import net.craftforge.essential.controller.annotations.*;
import net.craftforge.essential.controller.constants.HttpStatusCode;
import net.craftforge.essential.controller.constants.HttpMethod;
import net.craftforge.reflection.managers.ClassManager;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

/**
 * Utilities for annotation actions.
 *
 * @author Christian Bick
 * @since 17.02.2011
 */
public class AnnotationUtils {

    /**
     * Gets the @Path annotation value of a class, looking it up on type level. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The @Path annotation value, null if not found
     */
    public static String getPathFromClass(Class<?> clazz) {
        Path annotation = (Path) ClassManager.getInstance(clazz).getTypeLevelAnnotation(Path.class);
        return annotation == null ? null : ControllerUtils.standardUri(annotation.value());
    }

    /**
     * Gets the @Path annotation value of a method, looking it up on method level. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param method The method
     * @return The @Path annotation value, null if not found
     */
    public static String getPathFromMethod(Method method) {
        Path annotation = (Path) ClassManager.getInstance(method).getMethodLevelAnnotation(method, Path.class);
        return annotation == null ? null : ControllerUtils.standardUri(annotation.value());
    }

    /**
     * Gets the @Param annotations value of a field, looking it up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param field The field
     * @return The @Param annotation value, null if not found
     */
    public static String getParamFromProperty(Field field) {
        Param annotation = (Param) ClassManager.getInstance(field).getPropertyLevelAnnotation(field, Param.class);
        return annotation == null ? null : annotation.value();
    }

    /**
     * Gets the @DefaultValue annotation value(s) of a field, looking it up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param field The field
     * @return The @DefaultValue annotation value, null if not found
     */
    public static String[] getDefaultValuesFromProperty(Field field) {
        DefaultValue annotation =(DefaultValue) ClassManager.getInstance(field).getPropertyLevelAnnotation(field, DefaultValue.class);
        return annotation == null ? null : annotation.value();
    }

    /**
     * Gets the method of a producer being annotated with the @Produces annotation
     * whose value matches the specified accepted mime type.
     *
     * @param producer The producer method
     * @param accepted The accepted mime type
     * @return The method matching the annotation requirements
     * @throws net.craftforge.essential.controller.ControllerException Failed to find a producer method supplying the accepted mime type
     */
    public static Method getProducerMethod(Object producer, String accepted) throws ControllerException {
        int maxQuality = 0;
        Method selectedMethod = null;
        for (Method method : producer.getClass().getMethods()) {
            if (method.isAnnotationPresent(Produces.class)) {
                Produces annotation = method.getAnnotation(Produces.class);
                for (String produces : annotation.value()) {
                    int matchingQuality = getMatchingQuality(produces, accepted);
                    if (matchingQuality > maxQuality) {
                        selectedMethod = method;
                        maxQuality = matchingQuality;
                    }
                }
            }
        }
        if (selectedMethod == null) {
            throw new ControllerException("Producer class " + producer.getClass() + " does not supply any of the " +
                    "accepted mime types: " + accepted, HttpStatusCode.UnsupportedMediaType);
        }
        return selectedMethod;
    }

    /**
     * Gets the method of a consumer being annotated with the @Consumes annotation
     * whose value matches the specified mime type.
     *
     * @param consumer The consumer
     * @param mimeType The mime type to consume
     * @return The method matching the annotation requirements
     * @throws ControllerException Failed to find a consumer method supplying the mime type
     */
    public static Method getConsumerMethod(Object consumer, String mimeType) throws ControllerException {
        int maxQuality = 0;
        Method selectedMethod = null;
        for (Method method : consumer.getClass().getMethods()) {
            if (method.isAnnotationPresent(Consumes.class)) {
                Consumes annotation = method.getAnnotation(Consumes.class);
                for (String consumes : annotation.value()) {
                    int matchingQuality = getMatchingQuality(consumes, mimeType);
                    if (matchingQuality > maxQuality) {
                        selectedMethod = method;
                        maxQuality = matchingQuality;
                    }
                }
            }
        }
        if (selectedMethod == null) {
            throw new ControllerException("Given consumer does not supply the detected mime type: " + mimeType);
        }
        return selectedMethod;
    }

    /**
     * Gets the matching quality of an accepted and an available mime type enumeration.
     * The quality depends on how good the best of the available mime types matches against
     * the accepted mime type. The higher the number, the better is the quality.<br>
     * <br>
     * 0: None of the available mime types matches.<br>
     * 1: The best match was found at &#42;/&#42; vs. static/static<br>
     * 2: The best match was found at &#42;/&#42; vs. static/&#42;<br>
     * 3: The best match was found at &#42;/&#42; vs. &#42;/&#42;<br>
     * 4: The best match was fount at static/&#42; vs static/static<br>
     * 5: The best match was found at static/&#42; vs. static/&#42;<br>
     * 6: The best match was found at static/static vs. static/static (exact match)<br>
     *
     * @param availableEnum The comma separated enumeration of available mime types
     * @param acceptedEnum The comma separated enumeration of accepted mime types
     * @return The quality as a figure between 0 (lowest) and 6 (highest)
     * @throws ControllerException Invalid mime type declaration
     */
    private static int getMatchingQuality(String availableEnum, String acceptedEnum) throws ControllerException {
        int maxQuality = 0;
        for (String acceptedType : acceptedEnum.split(",")) {
            acceptedType = acceptedType.split(";")[0];
            acceptedType = acceptedType.trim();
            String[] acceptedParts = acceptedType.split("/");
            if (acceptedParts.length == 1) {
                acceptedParts = new String[] { acceptedParts[0], "*" };   
            }
            for (String availableType : availableEnum.split(",")) {
                availableType = availableType.trim();
                String[] availableParts = availableType.split("/");
                if (availableParts.length == 1) {
                    availableParts = new String[] { availableParts[0], "*" };   
                }
                int currentQuality = 0;
                if (availableParts[0].equals("*") && availableParts[1].equals("*")) {
                    if (acceptedParts[0].equals("*") && acceptedParts[1].equals("*")) {
                        currentQuality = 3;
                    } else if (acceptedParts[1].equals("*")) {
                        currentQuality = 2;
                    } else {
                        currentQuality = 1;
                    }
                } else if (availableParts[0].equals(acceptedParts[0]) && availableParts[1].equals("*")) {
                    if (acceptedParts[1].equals("*")) {
                        currentQuality = 5;
                    } else {
                        currentQuality = 4;
                    }
                } else if (availableParts[0].equals(acceptedParts[0]) && availableParts[1].equals(acceptedParts[1])) {
                    currentQuality = 6;
                }
                maxQuality = Math.max(currentQuality, maxQuality);
            }
        }
        return maxQuality;
    }
    /**
     * Gets the @Producer annotation value of a method, looking it up on method level first, then class level.
     * Uses the complete class hierarchy, including interfaces.
     *
     * @param method The method
     * @return The @Producer annotation value or null if not found
     */
    public static String getProducerFromMethodOrClass(Method method) {
        Producer annotation = (Producer) ClassManager.getInstance(method).getMethodOrTypeLevelAnnotation(method, Producer.class);
        return annotation == null ? null : ControllerUtils.standardUri(annotation.value());
    }

    /**
     * Gets the @Consumer annotation value of a method, looking it up on method level first, then class level.
     * Uses the complete class hierarchy, including interfaces.
     *
     * @param method The method
     * @return The @Consumer annotation value or null if not found
     */
    public static String getConsumerFromMethodOrClass(Method method) {
        Consumer annotation = (Consumer) ClassManager.getInstance(method).getMethodOrTypeLevelAnnotation(method, Consumer.class);
        return annotation == null ? null : ControllerUtils.standardUri(annotation.value());
    }

    /**
     * Checks if the @Public annotation is present, looking it up on method level first, then class level.
     * Uses the complete class hierarchy, including interfaces.
     *
     * @param method The method
     * @return Whether the @Public annotation is present or not
     */
    public static boolean isPublic(Method method) {
        Public annotation = (Public) ClassManager.getInstance(method).getMethodOrTypeLevelAnnotation(method, Public.class);
        return annotation != null;
    }

    /**
     * Gets a list of all fields with an @Param annotation, looking them up on field level
     * and on method level of getters and setters. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The list of fields
     */
    public static List<Field> getParamFieldsFromClass(Class<?> clazz) {
        return ClassManager.getInstance(clazz).getAllPropertiesAnnotatedWith(Param.class);
    }

    /**
     * Gets a list of all methods with a REST annotation (@Get, @Post, etc.)
     * in the class, looking it up on method level. Uses the complete class
     * hierarchy, including interfaces.
     *
     * @param clazz The class
     * @return The list of service methods
     */
    public static List<Method> getServiceMethodsFromClass(Class<?> clazz) {
        List<Method> methods = new ArrayList<Method>();
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Get.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Post.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Put.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Delete.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Head.class));
        methods.addAll(ClassManager.getInstance(clazz).getAllMethodsAnnotatedWith(Trace.class));
        return methods;
    }

    /**
     * Gets the REST annotation (@Get, @Post, etc.) of a method, looking it up on method
     * level. Uses the complete class hierarchy, including interfaces.
     *
     * @param method The method
     * @return The REST annotation or null if none found
     */
    public static String getHttpMethod(Method method) {
        Annotation annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Get.class);
        if (annotation != null) {
            return HttpMethod.GET;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Post.class);
        if (annotation != null) {
            return HttpMethod.POST;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Put.class);
        if (annotation != null) {
            return HttpMethod.PUT;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Delete.class);
        if (annotation != null) {
            return HttpMethod.DELETE;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Head.class);
        if (annotation != null) {
            return HttpMethod.HEAD;
        }
        annotation = ClassManager.getInstance(method).getMethodLevelAnnotation(method, Trace.class);
        if (annotation != null) {
            return HttpMethod.TRACE;
        }
        return null;
    }
}
