/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.utils;

import org.apache.commons.codec.binary.Base64;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.*;

/**
 * Utilities for controller actions.
 *
 * @author Christian Bick
 * @since 02.12.2010
 */
public class ControllerUtils {

    /**
     * Replaces all duplicate slashes of an URI and removes
     * leading as well as following slashes.
     *
     * @param uri The URI
     * @return The slimmed URI
     */
    public static String slimUri(String uri) {
        uri = uri.replaceAll("/+", "/");
        if (uri.endsWith("/")) {
            uri = uri.substring(0, uri.length()-1);
        }
        if (uri.startsWith("/")) {
            uri = uri.substring(1, uri.length());
        }
        return uri;
    }

    /**
     * Replaces all duplicate slashes of an URI,
     * removes all following slashes and ensures
     * a leading slash.
     *
     * @param uri The URI
     * @return The standardized URI
     */
    public static String standardUri(String uri) {
        uri = uri.replaceAll("/+", "/");
        if (uri.endsWith("/")) {
            uri = uri.substring(0, uri.length()-1);
        }
        if (! uri.startsWith("/")) {
            uri = "/" + uri;
        }
        return uri;
    }

    /**
     * Gets the parameters from an uri query part as a map-list combination.
     * Accepts params with same name and multiple values ("param=5&param=6") as well as array like
     * params ("param[0]=7&param[1]=8").
     *
     * @param encoding The URL encoding
     * @param uriQueryPart The query part of the uri without leading question mark
     * @return The parameters as a map-list combination
     * @throws UnsupportedEncodingException if encoding is not supported by URLDecoder
     */
    public static Map<String, String[]> getUrlParameters(String encoding, String uriQueryPart)
        throws UnsupportedEncodingException {

        uriQueryPart = uriQueryPart.replaceAll("\\[[0-9]+\\]", "");

        Map<String, List<String>> paramWorkingMap = new HashMap<String, List<String>>();
        for (String param : uriQueryPart.split("&")) {
            String pair[] = param.split("=");
            String paramName = URLDecoder.decode(pair[0], encoding);
            String paramValue = "";
            if (pair.length == 2) {
                 URLDecoder.decode(pair[1], encoding);
            }
            List<String> paramValues = paramWorkingMap.get(paramName);
            if (paramValues == null) {
                paramValues = new ArrayList<String>();
                paramWorkingMap.put(paramName, paramValues);
            }
            paramValues.add(paramValue);
        }

        Map<String, String[]> paramReturnMap = new HashMap<String, String[]>();
        for (String paramName : paramWorkingMap.keySet()) {
            List<String> paramValues = paramWorkingMap.get(paramName);
            String[] paramValueArray = new String[paramValues.size()];
            paramReturnMap.put(paramName, paramValues.toArray(paramValueArray));
        }
        return paramReturnMap;
    }

    /**
     * Gets an object from an XML output stream.
     *
     * @param clazz The object's class
     * @param outputStream The XML output stream
     * @return The object
     */
    public static Object getObjectFromXmlOutputStream(Class<?> clazz, ByteArrayOutputStream outputStream) {

        Object result = null;
        try {
            JAXBContext jaxbContext = JAXBContext.newInstance(clazz);
            javax.xml.bind.Unmarshaller unmarshaller = jaxbContext.createUnmarshaller();
            result = unmarshaller.unmarshal(new ByteArrayInputStream(outputStream.toByteArray()));
        } catch (JAXBException e) {
            e.printStackTrace();
        }
        return result;
    }

    /**
     * Gets an XML input stream from an object.
     *
     * @param object The object
     * @return The XML input stream
     */
    public static InputStream getXmlInputStreamFromObject(Object object) {
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        try {
            JAXBContext jaxbContext = JAXBContext.newInstance(object.getClass());
            Marshaller marshaller = jaxbContext.createMarshaller();
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            marshaller.marshal(object, outputStream);
        } catch (JAXBException e) {
            e.printStackTrace();
        }
        return new ByteArrayInputStream(outputStream.toByteArray());
    }

    /**
     * Gets an HTTP basic authorization header value from a username and password.
     *
     * @param username The username
     * @param password The password
     * @return The HTTP basic authorization header
     */
    public static String getBasicAuthorizationHeader(String username, String password) {
        String authPair = username + ":" + password;
        return "Basic " + new String(Base64.encodeBase64(authPair.getBytes()));
    }

    /**
     * Gets a username password pair from an HTTP authorization header value.
     *
     * @param authorizationHeader HTTP authorization header
     * @return The username password pair as Array of length 2
     */
    public static String[] getAuthPair(String authorizationHeader) {
        if (authorizationHeader == null || authorizationHeader.isEmpty()) {
            return new String[] { null, null };
        }

        StringTokenizer st = new StringTokenizer(authorizationHeader);
        String basic = st.nextToken();

        if (basic == null || ! basic.equalsIgnoreCase("Basic")) {
            return new String[] { null, null };
        }

        String credentials = st.nextToken();
        String userPass = new String(Base64.decodeBase64(credentials));
        int pivot = userPass.indexOf(":");
        if (pivot == -1 ) {
            return new String[] { null, null };
        }

        String user = userPass.substring(0, pivot);
        String pass = userPass.substring(pivot+1);

        return new String[] { user, pass };
    }
}
