/*
 * This file is part of essential (http://essential.craftforge.net).
 *
 *     Essential is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU Lesser Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     Essential is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (c) 2011 Christian Bick.
 */

package net.craftforge.essential.controller.utils;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Utilities for regular expression actions.
 *
 * @author Christian Bick
 * @since 17.02.2011
 */
public class RegExUtils {

    public static final String DEFAULT_REG_EX = "[^/]+?";

    public static final String OPEN_END_REG_EX = "(/.*)?";

    public static final Pattern FIND_REG_EX_OPERATORS = Pattern.compile("([\\\\*+\\[\\]()\\$.?\\^|])");

    public static final Pattern FIND_SLASH = Pattern.compile("/");

    public static final Pattern FIND_SEMI_COLON = Pattern.compile(";");

    public static final Pattern FIND_COLON = Pattern.compile(":");

    public static final Pattern FIND_ALL_LITERALS = Pattern.compile("[a-zA-Z0-9]");

    public static final Pattern FIND_ALL_NONE_LITERALS = Pattern.compile("[^a-zA-Z0-9]");

    public static final Pattern FIND_CURLY_BRACKETS = Pattern.compile("[\\{\\}]");

    public static final Pattern FIND_ENCLOSED_BY_CURLY_BRACKETS = Pattern.compile("\\{.*\\}");

    public static final Pattern FIND_GROUPS_WITH_SURROUNDING = Pattern.compile("[^(]*\\([^)]*\\)[^(]*");

    public static final Pattern FIND_NONE_DEFAULT_GROUPS_WITH_SURROUNDING = Pattern.compile("([^(]|\\(\\[\\^/\\]\\+\\?\\))*\\([^)]*\\)([^(]|\\(\\[\\^/\\]\\+\\?\\))*");

    public static final Pattern FIND_GROUPS = Pattern.compile("(\\^)*(\\([^)]*\\))(\\{.*\\})*([\\+\\*\\?\\$])*");

    /**
     * Transforms a path into a regular expression matching this path.
     *
     * @param path The path
     * @return The regular expression
     */
    public static String getRegExFromPath(String path) {
        path = ControllerUtils.slimUri(path);
        String[] uriParts = FIND_SLASH.split(path);
        String regEx = "";
        for (String part : uriParts) {
            if (FIND_ENCLOSED_BY_CURLY_BRACKETS.matcher(part).matches()) {
                if (FIND_COLON.matcher(part).find()) {
                    String[] paramParts = part.split(":");
                    regEx += "/(" + FIND_CURLY_BRACKETS.matcher(paramParts[1]).replaceAll("") + ")";
                } else {
                    regEx += "/(" + DEFAULT_REG_EX + ")";
                }
            } else {
                try {
                    regEx += new URI("/" + FIND_REG_EX_OPERATORS.matcher(part).replaceAll("\\\\$"));
                } catch (URISyntaxException e) {
                    throw new RuntimeException(e);
                }
            }
        }
        regEx += OPEN_END_REG_EX;
        return ControllerUtils.standardUri(regEx);
    }

    /**
     * Gets all values from a request path by splitting into different parts
     * using the slashes as separators. The parts are converted to an array
     * containing the values separated by semi colons.
     *
     * @param path The path
     * @return List, representing each request value
     */
    public static List<String[]> getValuesFromPath(String path) {
        path = ControllerUtils.slimUri(path);
        String[] values = FIND_SLASH.split(path);
        List<String[]> valuesArray = new ArrayList<String[]>();
        for (String value : values) {
            valuesArray.add(FIND_SEMI_COLON.split(value));
        }
        return valuesArray;
    }

    /**
     * Gets the path parameters from a path by splitting into different parts
     * using the slashes as separators.
     *
     * @param path The path
     * @return List, representing each path parameter
     */
    public static List<String> getParametersFromPath(String path) {
        path = ControllerUtils.slimUri(path);
        String[] uriParts = FIND_SLASH.split(path);
        List<String> params = new ArrayList<String>();
        for (String part : uriParts) {
            if (FIND_ENCLOSED_BY_CURLY_BRACKETS.matcher(part).matches()) {
                if (part.contains(":")) {
                    String[] paramParts = part.split(":");
                    params.add("{" + FIND_CURLY_BRACKETS.matcher(paramParts[0]).replaceAll("") + "}");
                } else {
                    params.add("{" + FIND_CURLY_BRACKETS.matcher(part).replaceAll("") + "}");
                }
            } else {
                params.add("{}");
            }
        }
        return params;
    }

    /**
     * Counts all groups except of the default groups being declared within the regular expression of a pattern.
     *
     * @param pattern The pattern
     * @return The amount of none-default groups
     */
    public static int countAllGroupsWithoutDefault(Pattern pattern) {
        String value = pattern.pattern();
        String result = FIND_NONE_DEFAULT_GROUPS_WITH_SURROUNDING.matcher(value).replaceAll("0");
        return result.length();
    }

    /**
     * Counts all groups being declared within the regular expression of a pattern.
     *
     * @param pattern The pattern
     * @return The amount of all groups
     */
    public static int countAllGroups(Pattern pattern) {
        String value = pattern.pattern();
        String result = FIND_GROUPS_WITH_SURROUNDING.matcher(value).replaceAll("0");
        return result.length();
    }

    /**
     * Counts all literals (text and figures)within the regular expression of pattern
     * that are not contained within a group.
     *
     * @param pattern The pattern
     * @return The amount of literals not contained in a group 
     */
    public static int countLiteralsWithoutGroups(Pattern pattern) {
        String value = pattern.pattern();
        value = FIND_GROUPS.matcher(value).replaceAll("");
        value = FIND_ALL_NONE_LITERALS.matcher(value).replaceAll("");
        return value.length();
    }
}
